import datetime
import logging
import os
import time

import numpy as np
import tensorflow as tf
from tensorflow.contrib import learn

from . import cnn_text_util


class CnnText(object):
    """
    A CNN for text classification.
    Uses an embedding layer, followed by a convolutional, max-pooling and softmax layer.
    This code is based on Denny Britz's CNN implementation for text classification in Tensorflow:
    http://www.wildml.com/2015/12/implementing-a-cnn-for-text-classification-in-tensorflow/
    """
    def __init__(
      self, sequence_length, num_classes, vocab_size,
      embedding_size, filter_sizes, num_filters, l2_reg_lambda=0.0):

        self.logger = logging.getLogger(__name__)

        # Placeholders for input, output and dropout
        self.input_x = tf.placeholder(tf.int32, [None, sequence_length], name="input_x")
        self.input_y = tf.placeholder(tf.float32, [None, num_classes], name="input_y")
        self.dropout_keep_prob = tf.placeholder(tf.float32, name="dropout_keep_prob")

        # Keeping track of l2 regularization loss (optional)
        l2_loss = tf.constant(0.0)

        # Embedding layer
        with tf.device('/cpu:0'), tf.name_scope("embedding"):
            self.W = tf.Variable(
                tf.random_uniform([vocab_size, embedding_size], -1.0, 1.0),
                name="W")
            self.embedded_chars = tf.nn.embedding_lookup(self.W, self.input_x)
            self.embedded_chars_expanded = tf.expand_dims(self.embedded_chars, -1)

        # Create a convolution + maxpool layer for each filter size
        pooled_outputs = []
        for i, filter_size in enumerate(filter_sizes):
            with tf.name_scope("conv-maxpool-%s" % filter_size):
                # Convolution Layer
                filter_shape = [filter_size, embedding_size, 1, num_filters]
                W = tf.Variable(tf.truncated_normal(filter_shape, stddev=0.1), name="W")
                b = tf.Variable(tf.constant(0.1, shape=[num_filters]), name="b")
                conv = tf.nn.conv2d(
                    self.embedded_chars_expanded,
                    W,
                    strides=[1, 1, 1, 1],
                    padding="VALID",
                    name="conv")
                # Apply nonlinearity
                h = tf.nn.relu(tf.nn.bias_add(conv, b), name="relu")
                # Maxpooling over the outputs
                pooled = tf.nn.max_pool(
                    h,
                    ksize=[1, sequence_length - filter_size + 1, 1, 1],
                    strides=[1, 1, 1, 1],
                    padding='VALID',
                    name="pool")
                pooled_outputs.append(pooled)

        # Combine all the pooled features
        num_filters_total = num_filters * len(filter_sizes)
        self.h_pool = tf.concat(pooled_outputs, 3)
        self.h_pool_flat = tf.reshape(self.h_pool, [-1, num_filters_total])

        # Add dropout
        with tf.name_scope("dropout"):
            self.h_drop = tf.nn.dropout(self.h_pool_flat, self.dropout_keep_prob)

        # Final (unnormalized) scores and predictions
        with tf.name_scope("output"):
            W = tf.get_variable(
                "W",
                shape=[num_filters_total, num_classes],
                initializer=tf.contrib.layers.xavier_initializer())
            b = tf.Variable(tf.constant(0.1, shape=[num_classes]), name="b")
            l2_loss += tf.nn.l2_loss(W)
            l2_loss += tf.nn.l2_loss(b)
            self.scores = tf.nn.xw_plus_b(self.h_drop, W, b, name="scores")
            self.predictions = tf.argmax(self.scores, 1, name="predictions")

        # CalculateMean cross-entropy loss
        with tf.name_scope("loss"):
            losses = tf.nn.softmax_cross_entropy_with_logits(logits=self.scores, labels=self.input_y)
            self.loss = tf.reduce_mean(losses) + l2_reg_lambda * l2_loss

        # Accuracy
        with tf.name_scope("accuracy"):
            correct_predictions = tf.equal(self.predictions, tf.argmax(self.input_y, 1))
            self.accuracy = tf.reduce_mean(tf.cast(correct_predictions, "float"), name="accuracy")


class CnnTextEvaluator(object):
    """
     CnnTextEvaluator
     """
    def __init__(self, cfg, current_category):
        self.logger = logging.getLogger(__name__)
        self.cfg = cfg
        self.current_category = current_category
        self.x_raw = None
        self.x_test = None
        # Map data into vocabulary
        self.vocab_path = os.path.join(self.cfg['pre_trained_dir'][self.current_category], "vocab")
        self.vocab_processor = learn.preprocessing.VocabularyProcessor.restore(self.vocab_path)

        self.sess = None
        self.predictions = None
        self.dropout_keep_prob = None
        self.input_x = None
        checkpoint_file = tf.train.latest_checkpoint(self.cfg['pre_trained_dir'][self.current_category] + "/checkpoints")
        graph = tf.Graph()
        with graph.as_default():
            session_conf = tf.ConfigProto(
                # device_count={'GPU':0},
                allow_soft_placement=self.cfg['allow_soft_placement'],
                log_device_placement=self.cfg['log_device_placement'])
            self.sess = tf.Session(config=session_conf)
            with self.sess.as_default():
                # Load the saved meta graph and restore variables
                self.logger.debug("Load the checkpoint: {}".format(checkpoint_file))
                saver = tf.train.import_meta_graph("{}.meta".format(checkpoint_file))
                saver.restore(self.sess, checkpoint_file)

                # Get the placeholders from the graph by name
                self.input_x = graph.get_operation_by_name("input_x").outputs[0]
                # input_y = graph.get_operation_by_name("input_y").outputs[0]
                self.dropout_keep_prob = graph.get_operation_by_name("dropout_keep_prob").outputs[0]

                # Tensors we want to evaluate
                self.predictions = graph.get_operation_by_name("output/predictions").outputs[0]

    def predict(self, x_raw=None):
        start = time.time()
        self.x_raw = x_raw
        self.x_test = np.array(list(self.vocab_processor.transform(self.x_raw)))

        # Generate batches for one epoch
        batches = cnn_text_util.batch_iter(list(self.x_test), self.cfg['batch_size'], 1, shuffle=False)

        # Collect the predictions here
        all_predictions = []

        for x_test_batch in batches:
            batch_predictions = self.sess.run(self.predictions,
                                              {self.input_x: x_test_batch, self.dropout_keep_prob: 1.0})
            all_predictions = np.concatenate([all_predictions, batch_predictions])

        response = [int(i) for i in all_predictions]
        end = time.time()
        self.logger.debug("Predict time: {} seconds".format(end - start))
        self.logger.debug("Response: {}".format(response))
        return response


class CnnTextTraining(object):
    """
     CnnTextTraining
     """
    def __init__(self, cfg):
        self.logger = logging.getLogger(__name__)
        self.logger.debug("CnnTextTraining init")
        self.cfg = cfg
        if self.cfg['word_embeddings']['default'] is not None:
            self.embedding_name = cfg['word_embeddings']['default']
            self.embedding_dimension = self.cfg['word_embeddings'][self.embedding_name]['dimension']
        else:
            self.embedding_dimension = self.cfg['word_embeddings']['embedding_dim']

    def fit(self, dataset, output):
        self.logger.debug("CnnTextTraining fit")

        x_text, y = dataset.load_data_labels()

        # Build vocabulary
        max_document_length = max([len(x.split(" ")) for x in x_text])
        vocab_processor = learn.preprocessing.VocabularyProcessor(max_document_length)
        x = np.array(list(vocab_processor.fit_transform(x_text)))

        # Randomly shuffle data
        np.random.seed(10)
        shuffle_indices = np.random.permutation(np.arange(len(y)))
        x_shuffled = x[shuffle_indices]
        y_shuffled = y[shuffle_indices]

        # Split train/test set
        # TODO: This is very crude, should use cross-validation
        dev_sample_index = -1 * int(self.cfg['dev_sample_percentage'] * float(len(y)))
        x_train, x_dev = x_shuffled[:dev_sample_index], x_shuffled[dev_sample_index:]
        y_train, y_dev = y_shuffled[:dev_sample_index], y_shuffled[dev_sample_index:]
        self.logger.debug("Vocabulary Size: {:d}".format(len(vocab_processor.vocabulary_)))
        self.logger.debug("Train/Dev split: {:d}/{:d}".format(len(y_train), len(y_dev)))

        with tf.Graph().as_default():
            session_conf = tf.ConfigProto(
                # device_count={'GPU':0},
                allow_soft_placement=self.cfg['allow_soft_placement'],
                log_device_placement=self.cfg['log_device_placement'])
            sess = tf.Session(config=session_conf)
            with sess.as_default():
                cnn = CnnText(
                    sequence_length=x_train.shape[1],
                    num_classes=y_train.shape[1],
                    vocab_size=len(vocab_processor.vocabulary_),
                    embedding_size=self.embedding_dimension,
                    filter_sizes=list(map(int, self.cfg['filter_sizes'].split(","))),
                    num_filters=self.cfg['num_filters'],
                    l2_reg_lambda=self.cfg['l2_reg_lambda'])

                # Define Training procedure
                global_step = tf.Variable(0, name="global_step", trainable=False)
                optimizer = tf.train.AdamOptimizer(1e-3)
                grads_and_vars = optimizer.compute_gradients(cnn.loss)
                train_op = optimizer.apply_gradients(grads_and_vars, global_step=global_step)

                # Keep track of gradient values and sparsity (optional)
                grad_summaries = []
                for g, v in grads_and_vars:
                    if g is not None:
                        grad_hist_summary = tf.summary.histogram("{}/grad/hist".format(v.name.replace(":","_")), g)
                        sparsity_summary = tf.summary.scalar("{}/grad/sparsity".format(v.name.replace(":","_")), tf.nn.zero_fraction(g))
                        grad_summaries.append(grad_hist_summary)
                        grad_summaries.append(sparsity_summary)
                grad_summaries_merged = tf.summary.merge(grad_summaries)

                # Output directory for models and summaries
                timestamp = str(int(time.time()))
                out_dir = os.path.abspath(os.path.join(os.path.curdir, output))
                self.logger.debug("Writing to {}\n".format(out_dir))

                # Summaries for loss and accuracy
                loss_summary = tf.summary.scalar("loss", cnn.loss)
                acc_summary = tf.summary.scalar("accuracy", cnn.accuracy)

                # Train Summaries
                train_summary_op = tf.summary.merge([loss_summary, acc_summary, grad_summaries_merged])
                train_summary_dir = os.path.join(out_dir, "summaries", "train")
                train_summary_writer = tf.summary.FileWriter(train_summary_dir, sess.graph)

                # Dev summaries
                dev_summary_op = tf.summary.merge([loss_summary, acc_summary])
                dev_summary_dir = os.path.join(out_dir, "summaries", "dev")
                dev_summary_writer = tf.summary.FileWriter(dev_summary_dir, sess.graph)

                # Checkpoint directory. Tensorflow assumes this directory already exists so we need to create it
                checkpoint_dir = os.path.abspath(os.path.join(out_dir, "checkpoints"))
                checkpoint_prefix = os.path.join(checkpoint_dir, "model")
                if not os.path.exists(checkpoint_dir):
                    os.makedirs(checkpoint_dir)
                saver = tf.train.Saver(tf.global_variables(), max_to_keep=self.cfg['num_checkpoints'])

                # Write vocabulary
                vocab_processor.save(os.path.join(out_dir, "vocab"))

                # Initialize all variables
                sess.run(tf.global_variables_initializer())
                if self.cfg['word_embeddings']['default'] is not None:
                    vocabulary = vocab_processor.vocabulary_
                    initW = None
                    if self.embedding_name == 'word2vec':
                        # load embedding vectors from the word2vec
                        self.logger.debug("Load word2vec file {}".format(self.cfg['word_embeddings']['word2vec']['path']))
                        initW = cnn_text_util.load_embedding_vectors_word2vec(vocabulary,
                                                                              self.cfg['word_embeddings']['word2vec']['path'],
                                                                              self.cfg['word_embeddings']['word2vec']['binary'])
                        self.logger.debug("word2vec file has been loaded")
                    elif self.embedding_name == 'glove':
                        # load embedding vectors from the glove
                        self.logger.debug("Load glove file {}".format(self.cfg['word_embeddings']['glove']['path']))
                        initW = cnn_text_util.load_embedding_vectors_glove(vocabulary,
                                                                           self.cfg['word_embeddings']['glove']['path'],
                                                                           self.embedding_dimension)
                        self.logger.debug("glove file has been loaded\n")
                    sess.run(cnn.W.assign(initW))

                def train_step(x_batch, y_batch):
                    """
                    A single training step
                    """
                    feed_dict = {
                        cnn.input_x: x_batch,
                        cnn.input_y: y_batch,
                        cnn.dropout_keep_prob: self.cfg['dropout_keep_prob']
                    }
                    _, step, summaries, loss, accuracy = sess.run(
                        [train_op, global_step, train_summary_op, cnn.loss, cnn.accuracy],
                        feed_dict)
                    time_str = datetime.datetime.now().isoformat()
                    self.logger.debug("{}: step {}, loss {:g}, acc {:g}".format(time_str, step, loss, accuracy))
                    train_summary_writer.add_summary(summaries, step)

                def dev_step(x_batch, y_batch, writer=None):
                    """
                    Evaluates model on a dev set
                    """
                    feed_dict = {
                        cnn.input_x: x_batch,
                        cnn.input_y: y_batch,
                        cnn.dropout_keep_prob: 1.0
                    }
                    step, summaries, loss, accuracy = sess.run(
                        [global_step, dev_summary_op, cnn.loss, cnn.accuracy],
                        feed_dict)
                    time_str = datetime.datetime.now().isoformat()
                    self.logger.debug("{}: step {}, loss {:g}, acc {:g}".format(time_str, step, loss, accuracy))
                    if writer:
                        writer.add_summary(summaries, step)

                # Generate batches
                batches = cnn_text_util.batch_iter(
                    list(zip(x_train, y_train)), self.cfg['batch_size'], self.cfg['num_epochs'])
                # Training loop. For each batch...
                for batch in batches:
                    x_batch, y_batch = zip(*batch)
                    train_step(x_batch, y_batch)
                    current_step = tf.train.global_step(sess, global_step)
                    if current_step % self.cfg['evaluate_every'] == 0:
                        self.logger.debug("\nEvaluation:")
                        dev_step(x_dev, y_dev, writer=dev_summary_writer)
                        self.logger.debug("")
                    if current_step % self.cfg['checkpoint_every'] == 0:
                        path = saver.save(sess, checkpoint_prefix, global_step=current_step)
                        self.logger.debug("Saved model checkpoint to {}\n".format(path))



import numpy as np
import re
from sklearn.datasets import fetch_20newsgroups
from sklearn.datasets import load_files


def clean_str(string):
    """
    Tokenization/string cleaning for all datasets except for SST.
    Original taken from https://github.com/yoonkim/CNN_sentence/blob/master/process_data.py
    """
    string = re.sub(r"[^A-Za-z0-9(),!?\'\`]", " ", string)
    string = re.sub(r"\'s", " \'s", string)
    string = re.sub(r"\'ve", " \'ve", string)
    string = re.sub(r"n\'t", " n\'t", string)
    string = re.sub(r"\'re", " \'re", string)
    string = re.sub(r"\'d", " \'d", string)
    string = re.sub(r"\'ll", " \'ll", string)
    string = re.sub(r",", " , ", string)
    string = re.sub(r"!", " ! ", string)
    string = re.sub(r"\(", " \( ", string)
    string = re.sub(r"\)", " \) ", string)
    string = re.sub(r"\?", " \? ", string)
    string = re.sub(r"\s{2,}", " ", string)
    return string.strip().lower()


def batch_iter(data, batch_size, num_epochs, shuffle=True):
    """
    Generates a batch iterator for a dataset.
    """
    data = np.array(data)
    data_size = len(data)
    num_batches_per_epoch = int((len(data)-1)/batch_size) + 1
    for epoch in range(num_epochs):
        # Shuffle the data at each epoch
        if shuffle:
            shuffle_indices = np.random.permutation(np.arange(data_size))
            shuffled_data = data[shuffle_indices]
        else:
            shuffled_data = data
        for batch_num in range(num_batches_per_epoch):
            start_index = batch_num * batch_size
            end_index = min((batch_num + 1) * batch_size, data_size)
            yield shuffled_data[start_index:end_index]


def get_datasets_20newsgroup(subset='train', categories=None, shuffle=True, random_state=42):
    """
    Retrieve data from 20 newsgroups
    :param subset: train, test or all
    :param categories: List of newsgroup name
    :param shuffle: shuffle the list or not
    :param random_state: seed integer to shuffle the dataset
    :return: data and labels of the newsgroup
    """
    datasets = fetch_20newsgroups(subset=subset, categories=categories, shuffle=shuffle, random_state=random_state)
    return datasets


def get_datasets_mrpolarity(positive_data_file, negative_data_file):
    """
    Loads MR polarity data from files, splits the data into words and generates labels.
    Returns split sentences and labels.
    """
    # Load data from files
    positive_examples = list(open(positive_data_file, "r").readlines())
    positive_examples = [s.strip() for s in positive_examples]
    negative_examples = list(open(negative_data_file, "r").readlines())
    negative_examples = [s.strip() for s in negative_examples]

    datasets = dict()
    datasets['data'] = positive_examples + negative_examples
    target = [0 for x in positive_examples] + [1 for x in negative_examples]
    datasets['target'] = target
    datasets['target_names'] = ['positive_examples', 'negative_examples']
    return datasets


def get_datasets_localdata(container_path=None, categories=None, load_content=True,
                       encoding='utf-8', shuffle=True, random_state=42):
    """
    Load text files with categories as subfolder names.
    Individual samples are assumed to be files stored a two levels folder structure.
    :param container_path: The path of the container
    :param categories: List of classes to choose, all classes are chosen by default (if empty or omitted)
    :param shuffle: shuffle the list or not
    :param random_state: seed integer to shuffle the dataset
    :return: data and labels of the dataset
    """
    datasets = load_files(container_path=container_path, categories=categories,
                          load_content=load_content, shuffle=shuffle, encoding=encoding,
                          random_state=random_state)
    return datasets


def load_data_labels(datasets):
    """
    Load data and labels
    :param datasets:
    :return:
    """
    # Split by words
    x_text = datasets['data']
    x_text = [clean_str(sent) for sent in x_text]
    # Generate labels
    labels = []
    for i in range(len(x_text)):
        label = [0 for j in datasets['target_names']]
        label[datasets['target'][i]] = 1
        labels.append(label)
    y = np.array(labels)
    return [x_text, y]


def load_embedding_vectors_word2vec(vocabulary, filename, binary):
    # load embedding_vectors from the word2vec
    encoding = 'utf-8'
    with open(filename, "rb") as f:
        header = f.readline()
        vocab_size, vector_size = map(int, header.split())
        # initial matrix with random uniform
        embedding_vectors = np.random.uniform(-0.25, 0.25, (len(vocabulary), vector_size))
        if binary:
            binary_len = np.dtype('float32').itemsize * vector_size
            for line_no in range(vocab_size):
                word = []
                while True:
                    ch = f.read(1)
                    if ch == b' ':
                        break
                    if ch == b'':
                        raise EOFError("unexpected end of input; is count incorrect or file otherwise damaged?")
                    if ch != b'\n':
                        word.append(ch)
                word = str(b''.join(word), encoding=encoding, errors='strict')
                idx = vocabulary.get(word)
                if idx != 0:
                    embedding_vectors[idx] = np.fromstring(f.read(binary_len), dtype='float32')
                else:
                    f.seek(binary_len, 1)
        else:
            for line_no in range(vocab_size):
                line = f.readline()
                if line == b'':
                    raise EOFError("unexpected end of input; is count incorrect or file otherwise damaged?")
                parts = str(line.rstrip(), encoding=encoding, errors='strict').split(" ")
                if len(parts) != vector_size + 1:
                    raise ValueError("invalid vector on line %s (is this really the text format?)" % (line_no))
                word, vector = parts[0], list(map('float32', parts[1:]))
                idx = vocabulary.get(word)
                if idx != 0:
                    embedding_vectors[idx] = vector
        f.close()
        return embedding_vectors


def load_embedding_vectors_glove(vocabulary, filename, vector_size):
    # load embedding_vectors from the glove
    # initial matrix with random uniform
    embedding_vectors = np.random.uniform(-0.25, 0.25, (len(vocabulary), vector_size))
    f = open(filename)
    for line in f:
        values = line.split()
        word = values[0]
        vector = np.asarray(values[1:], dtype="float32")
        idx = vocabulary.get(word)
        if idx != 0:
            embedding_vectors[idx] = vector
    f.close()
    return embedding_vectors



from sklearn.datasets import load_files
from . import Dataset


class Generic(Dataset):

    def __init__(self, cfg=None):
        """
        Load text files with categories as subfolder names.
        Individual samples are assumed to be files stored a two levels folder structure.
        :param container_path: The path of the container
        :param categories: List of classes to choose, all classes are chosen by default (if empty or omitted)
        :param shuffle: shuffle the list or not
        :param random_state: seed integer to shuffle the dataset
        :return: data and labels of the dataset
        """
        super().__init__()
        self.__dataset__ = load_files(container_path=cfg['container_path'], categories=cfg['categories'],
                                      load_content=cfg['load_content'], shuffle=cfg['shuffle'],
                                      encoding=cfg['encoding'], random_state=cfg['random_state'])



import sys
import logging.config
import threading
import socketserver
import socket
import hashlib
import struct
import json
import logging
from opentc.core import __version__


class Server(object):
    """
    Class for using Server with a network socket
    """
    classifiers = dict()
    port = None

    def __init__(self, cfg=None):
        """
        class initialisation
        """
        self.logger = logging.getLogger(__name__)
        self.cfg = cfg
        self.server = None
        self.address = None
        self.port = None
        self.timeout = None
        for classifier_name in self.cfg['classifiers']:
            if classifier_name == "default":
                continue
            module = __import__("opentc.core.classifier." + classifier_name)
            class_ = getattr(getattr(getattr(getattr(module, "core"), "classifier"),
                                     classifier_name), classifier_name.title())
            # class_ = attr4
            if class_ is not None:
                classifier = dict()
                classifier['enabled'] = self.cfg['classifiers'][classifier_name]['enabled']
                dataset_name = self.cfg['dataset']['name']
                classifier['class'] = class_(self.cfg['classifiers'][classifier_name],
                                             self.cfg['dataset']['categories'],
                                             dataset_name)
                Server.classifiers[classifier_name] = classifier

    def start(self, address=None, port=None, timeout=None, run_forever=True):
        """
        :param run_forever:
        :param address: hostname or ip address
        :param port: TCP port
        :param timeout: socket timeout
        :return:
        """
        if address:
            self.address = address
        else:
            self.address = self.cfg["address"]
        if port:
            self.port = port
        else:
            self.port = self.cfg["port"]
        if timeout:
            self.timeout = timeout
        else:
            self.timeout = self.cfg["timeout"]

        try:
            self.server = self.ThreadedTCPServer((self.address, self.port),
                                                 self.ThreadedTCPRequestHandler)
            self.server.socket.settimeout(self.timeout)
            self.server.socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

            # Start a thread with the server -- that thread will then start one
            # more thread for each request
            server_thread = threading.Thread(target=self.server.serve_forever)
            # Exit the server thread when the main thread terminates
            server_thread.daemon = True
            server_thread.start()
            self.logger.info("Server loop running in thread: {}".format(server_thread.name))
            if run_forever:
                self.server.serve_forever()
        except socket.error:
            e = sys.exc_info()[1]
            raise ConnectionError(e)

    def shutdown(self):
        self.logger.info("Server shutdown")
        self.server.shutdown()

    class ThreadedTCPRequestHandler(socketserver.StreamRequestHandler):
        max_buffer_size = 4096

        def __init__(self, request, client_address, server):
            self.logger = logging.getLogger(__name__)
            super().__init__(request, client_address, server)

        def handle(self):

            try:
                cur_thread = threading.current_thread()
                while True:
                    data = self.receive()
                    if data is None:
                        break
                    data = data.rstrip()
                    self.logger.info("Thread {} received: {}".format(cur_thread.name, data))
                    header = data.split(b':')
                    if header[0] == b'PING':
                        self.ping(header[1].decode('utf-8'))
                    elif header[0] == b'VERSION':
                        self.version()
                    elif header[0] == b'RELOAD':
                        self.reload()
                    elif header[0] == b'LIST_CLASSIFIER':
                        self.list_classifier()
                    elif header[0] == b'SET_CLASSIFIER':
                        classifier = header[1].decode('utf-8')
                        value = header[2].decode('utf-8')
                        self.set_classifier(classifier, value)
                    elif header[0] == b'MD5_FILE':
                        file_name = header[1]
                        self.md5_file(file_name=file_name)
                    elif header[0] == b'MD5_STREAM':
                        self.md5_stream()
                    elif header[0] == b'PREDICT_STREAM':
                        self.predict_stream()
                    elif header[0] == b'PREDICT_FILE':
                        file_name = header[1]
                        self.predict_file(file_name=file_name)
                    elif header[0] == b'CLOSE':
                        self.close()
                        break
                    else:
                        self.unknown_command()
            except socket.error:
                e = sys.exc_info()[1]
                raise ConnectionError(e)
            self.logger.info("Thread {} exit".format(cur_thread.name))

        def send(self, data):
            size = len(data)
            packed_header = struct.pack('=I', size)
            self.request.sendall(packed_header + data)

        def receive(self):
            packed_header = self.rfile.read(4)
            if packed_header == b'':
                return None
            (size, ) = struct.unpack('=I', packed_header)
            if size == 0 or size > self.max_buffer_size:
                return None
            data = self.rfile.read(size)
            return data

        def ping(self, mid=0):
            response = dict()
            response["status"] = "OK"
            response["mid"] = mid
            response["result"] = "PONG"
            response = json.dumps(response).encode('utf-8')
            self.send(response)
            self.logger.debug("Ping sent pong with mid: {}".format(mid))

        def close(self):
            response = dict()
            response["status"] = "OK"
            response["result"] = "Bye"
            response = json.dumps(response).encode('utf-8')
            self.send(response)

        def version(self):
            response = dict()
            response["status"] = "OK"
            response["result"] = __version__
            response = json.dumps(response).encode('utf-8')
            self.send(response)

        def reload(self):
            response = b'reload'
            self.send(response)

        def list_classifier(self):
            response = dict()
            response["status"] = "OK"
            response["result"] = {classifier: Server.classifiers[classifier]['enabled']
                                  for classifier in Server.classifiers}
            response = json.dumps(response).encode('utf-8')
            self.send(response)

        def set_classifier(self, classifier, value):
            response = dict()
            response["status"] = "OK"
            if value.lower() == "true":
                Server.classifiers[classifier]['enabled'] = True
            elif value.lower() == "false":
                Server.classifiers[classifier]['enabled'] = False
            else:
                response["status"] = "ERROR"
            response["result"] = [{classifier: Server.classifiers[classifier]['enabled']}]
            response = json.dumps(response).encode('utf-8')
            self.send(response)

        def md5_file(self, file_name=None):
            # This function is just for testing purpose
            hash_md5 = hashlib.md5(open(file_name, 'rb').read())
            response = dict()
            if hash_md5:
                response["status"] = "OK"
                response["result"] = hash_md5.hexdigest()
                response = json.dumps(response).encode('utf-8')
            else:
                response["status"] = "Error"
                response["result"] = ""
                response = json.dumps(response).encode('utf-8')
            self.send(response)

        def md5_stream(self):
            # This function is just for testing purpose
            hash_md5 = hashlib.md5()
            while True:
                data = self.receive()
                if data is None:
                    break
                hash_md5.update(data)
            response = dict()
            if hash_md5:
                response["status"] = "OK"
                response["result"] = hash_md5.hexdigest()
                response = json.dumps(response).encode('utf-8')
            else:
                response["status"] = "Error"
                response["result"] = ""
                response = json.dumps(response).encode('utf-8')
            self.send(response)

        def predict_stream(self):
            stream = b''
            while True:
                data = self.receive()
                self.logger.debug("Data: {}".format(data))
                if data is None:
                    break
                stream += data
            stream = stream.decode('utf-8')
            multi_line = stream.split('\n')
            response = dict()
            response["status"] = "OK"
            result = dict()
            for classifier_name in Server.classifiers.keys():
                if Server.classifiers[classifier_name]['enabled']:
                    result[classifier_name] = \
                        Server.classifiers[classifier_name]['class'].predict(multi_line)
            response["result"] = result
            response = json.dumps(response).encode('utf-8')
            self.send(response)

        def predict_file(self, file_name=None):
            data = open(file_name, 'rb').read().decode('utf-8')
            multi_line = data.split('\n')
            response = dict()
            response["status"] = "OK"
            result = dict()
            for classifier_name in Server.classifiers.keys():
                if Server.classifiers[classifier_name]['enabled']:
                    result[classifier_name] = \
                        Server.classifiers[classifier_name]['class'].predict(multi_line)
            response["result"] = result
            response = json.dumps(response).encode('utf-8')
            self.send(response)

        def unknown_command(self):
            response = dict()
            response["status"] = "ERROR"
            response["result"] = "Unknown Command"
            response = json.dumps(response).encode('utf-8')
            self.send(response)

    class ThreadedTCPServer(socketserver.ThreadingMixIn, socketserver.TCPServer):
        allow_reuse_address = True
        pass


from . import Dataset


class Mrpolarity(Dataset):

    def __init__(self, cfg=None):
        """
        Loads MR polarity data from files, splits the data into words and generates labels.
        Returns split sentences and labels.
        """
        super().__init__()
        # Load data from files
        positive_examples = []
        negative_examples = []
        try:
            positive_examples = list(open(cfg['positive_data_file']['path'], "r").readlines())
            positive_examples = [s.strip() for s in positive_examples]
            negative_examples = list(open(cfg['negative_data_file']['path'], "r").readlines())
            negative_examples = [s.strip() for s in negative_examples]
        except OSError as err:
            self.logger.error("OS error: {0}".format(err))
            exit(1)

        self.__dataset__ = dict()
        self.__dataset__['data'] = positive_examples + negative_examples
        target = [0 for x in positive_examples] + [1 for x in negative_examples]
        self.__dataset__['target'] = target
        self.__dataset__['target_names'] = ['positive_examples', 'negative_examples']



"""
OpenTC is text classification engine running as client server architecture. It listen on port 3333

"""
from opentc.core.setting import __version__


__author__ = "Cahya Wirawan <Cahya.Wirawan@gmail.com>"
